"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var main_1 = require("../main");
var operators_1 = require("../operators");
var scopes_1 = require("../scopes");
var test_data_1 = require("./test-data/test-data");
describe('transform()', function () {
    it('[omit] should remove the specified properties from the results', function () {
        var results = (0, main_1.mocklify)()
            .add(2, test_data_1.MOCK_USERS)
            .transform((0, operators_1.omit)(['isAdmin', 'isOnline']))
            .getAll();
        expect(results).toEqual([
            Object.assign({}, test_data_1.MOCK_USERS[0], { isAdmin: undefined, isOnline: undefined }),
            Object.assign({}, test_data_1.MOCK_USERS[1], { isAdmin: undefined, isOnline: undefined }),
        ]);
    });
    it('[omit] should have no effect if no properties are provided', function () {
        var results = (0, main_1.mocklify)()
            .addAll(test_data_1.MOCK_USERS)
            .transform((0, operators_1.omit)([]))
            .getAll();
        expect(results).toEqual(test_data_1.MOCK_USERS);
    });
    it('[override] allows overriding specific props of results', function () {
        var results = (0, main_1.mocklify)()
            .addAll(test_data_1.MOCK_USERS)
            .transform((0, operators_1.override)({ points: 99 }))
            .getAll();
        expect(results.every(function (user) { return user.points === 99; }));
    });
    it('[modify] allows immutable modification of results', function () {
        var results = (0, main_1.mocklify)()
            .add(2, test_data_1.MOCK_USERS)
            .transform((0, operators_1.modify)(function (user, index) {
            user.id = "user_".concat(index);
            user.points *= 2;
        }))
            .getAll();
        expect(results).toEqual([
            Object.assign({}, test_data_1.MOCK_USERS[0], { id: "user_0", points: test_data_1.MOCK_USERS[0].points * 2 }),
            Object.assign({}, test_data_1.MOCK_USERS[1], { id: "user_1", points: test_data_1.MOCK_USERS[1].points * 2 }),
        ]);
    });
    it('[modify] modify callback provides the current item, its index and the full array of items', function () {
        var results = (0, main_1.mocklify)()
            .add(2, test_data_1.MOCK_USERS)
            .transform((0, operators_1.modify)(function (user, index, allUsers) {
            user.firstName = "".concat(user.firstName, " (user ").concat(index + 1, " of ").concat(allUsers.length, ")");
        }))
            .getAll();
        expect(results).toEqual([
            Object.assign({}, test_data_1.MOCK_USERS[0], { firstName: "".concat(test_data_1.MOCK_USERS[0].firstName, " (user 1 of 2)") }),
            Object.assign({}, test_data_1.MOCK_USERS[1], { firstName: "".concat(test_data_1.MOCK_USERS[1].firstName, " (user 2 of 2)") }),
        ]);
    });
    it('[omit > modify > override] allows multiple operators to be applied as a chain', function () {
        var results = (0, main_1.mocklify)()
            .add(1, test_data_1.MOCK_USERS)
            .transform((0, operators_1.omit)(['isOnline']), (0, operators_1.modify)(function (user, index) {
            user.id = "user_".concat(index);
            user.firstName = user.firstName.charAt(0).toUpperCase();
        }), (0, operators_1.override)({ isAdmin: true }))
            .getAll();
        expect(results).toEqual([
            Object.assign({}, test_data_1.MOCK_USERS[0], {
                isOnline: undefined,
                id: "user_0",
                firstName: test_data_1.MOCK_USERS[0].firstName.charAt(0).toUpperCase(),
                isAdmin: true
            })
        ]);
    });
    it('[override > modify] applies operators in the correct order', function () {
        var overrideThenModify = (0, main_1.mocklify)()
            .add(1, test_data_1.MOCK_USERS)
            .transform((0, operators_1.override)({ points: 20 }), (0, operators_1.modify)(function (user) { return user.points *= 10; }))
            .getAll();
        var modifyThenOverride = (0, main_1.mocklify)()
            .add(1, test_data_1.MOCK_USERS)
            .transform((0, operators_1.modify)(function (user) { return user.points *= 10; }), (0, operators_1.override)({ points: 20 }))
            .getAll();
        expect(overrideThenModify).toEqual([
            Object.assign({}, test_data_1.MOCK_USERS[0], { points: 20 * 10 })
        ]);
        expect(modifyThenOverride).toEqual([
            Object.assign({}, test_data_1.MOCK_USERS[0], { points: 20 })
        ]);
    });
    it('[omit > ~override > ~modify > ~omit] operators can be restricted to apply only on a subset of items', function () {
        var isGryffindor = function (user) { return user.tagIds.includes(test_data_1.MOCK_TAGS.gryffindor.id); };
        var isSlytherin = function (user) { return user.tagIds.includes(test_data_1.MOCK_TAGS.slytherin.id); };
        var isHarry = function (user) { return user.firstName === 'Harry' && user.lastName === 'Potter'; };
        var results = (0, main_1.mocklify)()
            .addAll(test_data_1.MOCK_USERS)
            .transform((0, operators_1.omit)(['isOnline']), (0, scopes_1.where)(isGryffindor, (0, operators_1.override)({ points: 1000 })), (0, scopes_1.where)(isSlytherin, (0, operators_1.modify)(function (user) { return user.points = 0; })), (0, scopes_1.where)(isHarry, (0, operators_1.override)({
            points: 9000,
            isAdmin: true
        })))
            .getAll();
        expect(results).toEqual(test_data_1.MOCK_USERS.map(function (user, index) {
            var expectedPoints = user.points;
            var expectedIsAdmin = user.isAdmin;
            if (isGryffindor(user, index)) {
                expectedPoints = 1000;
            }
            if (isSlytherin(user, index)) {
                expectedPoints = 0;
            }
            if (isHarry(user, index)) {
                expectedPoints = 9000;
                expectedIsAdmin = true;
            }
            return Object.assign({}, user, {
                isOnline: undefined,
                points: expectedPoints,
                isAdmin: expectedIsAdmin
            });
        }));
    });
});
